from __future__ import annotations

import pytest
from ase.build import bulk
from asedb import AtomsModel, Element


@pytest.fixture
def li_atoms():
    return bulk("Li") * (2, 2, 2)


def test_element(li_atoms, session):
    atoms = li_atoms
    hits = session.query(Element).all()
    assert len(hits) == 0

    model = AtomsModel.from_atoms(atoms)
    session.add(model)
    session.commit()

    hits = session.query(Element).all()
    assert len(hits) == 1
    assert hits[0].symbol == "Li"
    assert hits[0].count == len(atoms)
    first_id = hits[0].id

    atoms.symbols = "Na"

    # Change the atoms, verify we also update the Element object automatically
    model.set_atoms(atoms)
    session.add(model)
    session.commit()

    hits = session.query(Element).all()
    assert len(hits) == 1
    assert hits[0].symbol == "Na"
    assert hits[0].count == len(atoms)
    assert hits[0].id == first_id + 1



from .version import __version__
from .atoms_model import AtomsModel, Element, Calculation
from .initialization import initialize_engine, make_sqlite_engine
from .trajectory_model import Trajectory

__all__ = [
    "AtomsModel",
    "Element",
    "Calculation",
    "Trajectory",
    "initialize_engine",
    "make_sqlite_engine",
]



from __future__ import annotations

import ase
import numpy as np
from ase.build import bulk, fcc111
from asedb.atoms_model import AtomsModel
from asedb.trajectory_model import Trajectory


def make_atoms_list():
    return [
        ase.Atoms("H2", positions=[[0, 0, 0], [0, 0, 0.7]]),
        bulk("NaCl", crystalstructure="rocksalt", a=4.0).repeat((3, 2, 2)),
        fcc111("Al", size=(2, 2, 3), vacuum=10.0),
    ]


def test_trajectory(session):
    traj = Trajectory()
    expect = make_atoms_list()
    for atoms in expect:
        traj.add_atoms(atoms)
    session.add(traj)
    session.commit()

    loaded: Trajectory = session.query(Trajectory).filter_by(id=1).scalar()

    assert len(loaded.atoms_list) == 3
    atoms_list = loaded.to_atoms_list()
    assert len(atoms_list) == 3
    for ii, atoms in enumerate(atoms_list):
        assert isinstance(atoms, ase.Atoms)
        exp = expect[ii]
        assert atoms is not exp
        assert len(atoms) == len(exp)
        assert (atoms.symbols == exp.symbols).all()
        assert np.allclose(atoms.positions, exp.positions)


def test_trajectory_delete_cascade(session):
    traj = Trajectory()
    expect = make_atoms_list()
    for atoms in expect:
        traj.add_atoms(atoms)
    session.add(traj)
    session.commit()

    hits = session.query(AtomsModel).all()
    assert len(hits) == 3

    session.expire_all()
    del traj.atoms_list[0]
    session.add(traj)
    session.commit()

    hits = session.query(AtomsModel).all()
    assert len(hits) == 2



from __future__ import annotations

import random

import numpy as np
import pytest
from ase.build import bulk
from ase.calculators.singlepoint import SinglePointCalculator
from asedb.atoms_model import AtomsModel, CalcArray, Calculation
from asedb.properties import ArrayProperties


@pytest.fixture
def atoms():
    atoms = bulk("Au", cubic=True) * (3, 3, 2)
    energy = random.uniform(-10, 10)
    forces = np.random.uniform(-2, 2, size=(len(atoms), 3))
    calc = SinglePointCalculator(atoms, energy=energy, forces=forces)
    atoms.calc = calc
    return atoms


def test_calculation(atoms, session):
    hits = session.query(Calculation).all()
    assert len(hits) == 0

    session.add(AtomsModel.from_atoms(atoms))
    session.commit()

    hits = session.query(CalcArray).all()
    assert len(hits) == 1

    hits = session.query(Calculation).all()
    assert len(hits) == 1

    calc_model = hits[0]
    assert len(calc_model.arrays) == 1
    assert calc_model.arrays[0].name == ArrayProperties.FORCES
    calc_model.drop_array(ArrayProperties.FORCES)
    session.add(calc_model)
    session.commit()

    del calc_model

    hits = session.query(Calculation).all()
    assert len(hits) == 1
    loaded = hits[0]
    assert len(loaded.arrays) == 0

    hits = session.query(CalcArray).all()
    assert len(hits) == 0


def test_calc_array(atoms, session):
    session.add(AtomsModel.from_atoms(atoms))
    session.commit()

    hits = session.query(CalcArray).all()
    assert len(hits) == 1
    model = hits[0]
    assert isinstance(model.array_obj, np.ndarray)

    assert "CalcArray" in repr(model)

    array = model.get_array()
    assert isinstance(array, np.ndarray)
    assert array.dtype == atoms.get_forces().dtype
    assert np.allclose(array, atoms.get_forces())



from __future__ import annotations

import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy.schema import CreateSchema

from .abstract import Base


def make_sqlite_engine(filename: str, initialize: bool = False) -> sa.Engine:
    """Helper function to create a sqlite Engine,
    that disables the usage of the default schema."""
    engine = create_engine(
        f"sqlite:///{filename}",
        execution_options={"schema_translate_map": {"asedb": None}},
    )
    if initialize:
        initialize_engine(engine)
    return engine


def initialize_engine(engine: sa.Engine):
    dialect = engine.dialect.name
    if dialect != "sqlite":
        # sqlite doesn't support schemas
        create_schema(engine)
    Base.metadata.create_all(engine)


def create_schema(engine: sa.Engine):
    with engine.begin() as con:
        con.execute(CreateSchema("asedb", if_not_exists=True))
        con.commit()



from __future__ import annotations

import numpy as np
import pytest
from asedb.abstract import NamedArray


class DummyNamedArray(NamedArray):
    """Test version of a NamedArray"""

    __tablename__ = "test_table"


@pytest.mark.parametrize(
    "arr",
    [
        np.array([1, 2, 3]),
        np.array([1.1, 2.2, 3.3]),
    ],
)
def test_named_array(arr, session):
    model = DummyNamedArray.from_np_array("dummy", arr)
    session.add(model)
    session.commit()

    id = model.id
    del model
    old_arr = arr.copy()
    arr *= 20

    loaded = session.query(DummyNamedArray).filter_by(id=id).scalar()
    loaded_arr = loaded.get_array()
    assert isinstance(loaded_arr, np.ndarray)
    assert np.allclose(loaded_arr, old_arr)
    assert not np.allclose(loaded_arr, arr)


def test_set_array(session):
    arr = np.array([1, 2, 3])
    model = DummyNamedArray.from_np_array("dummy", arr)

    session.add(model)
    session.commit()

    assert np.allclose(model.get_array(), arr)

    new_arr = np.array([1.2, 1.5, 200.4, 3])
    model.set_array(new_arr)
    session.add(model)
    session.commit()

    loaded_model = session.query(DummyNamedArray).first()
    loaded = loaded_model.get_array()

    assert loaded.shape != arr.shape
    assert loaded.dtype != arr.dtype
    assert loaded.shape == new_arr.shape
    assert loaded.dtype == new_arr.dtype
    assert np.allclose(loaded, new_arr)



from __future__ import annotations

import itertools
import random

import ase
import numpy as np
import pytest
from ase.build import bulk, fcc111
from ase.calculators.singlepoint import SinglePointCalculator
from asedb import AtomsModel, Element
from asedb.atoms_model import Calculation, _decode_pbc, _encode_pbc


def make_atoms_list():
    return [
        ase.Atoms("H2", positions=[[0, 0, 0], [0, 0, 0.7]]),
        bulk("NaCl", crystalstructure="rocksalt", a=4.0).repeat((3, 2, 2)),
        fcc111("Al", size=(2, 2, 3), vacuum=10.0),
    ]


def cmp_atoms(atoms1: ase.Atoms, atoms2: ase.Atoms) -> None:
    assert atoms2 is not atoms1
    assert isinstance(atoms1, ase.Atoms)
    assert isinstance(atoms2, ase.Atoms)
    assert len(atoms2) == len(atoms1)
    assert np.equal(atoms1.numbers, atoms2.numbers).all()
    assert (atoms2.symbols == atoms1.symbols).all()
    assert np.allclose(atoms1.get_positions(), atoms2.positions)
    assert (atoms1.pbc == atoms2.pbc).all()
    assert (atoms1.get_tags() == atoms2.get_tags()).all()
    assert np.allclose(atoms1.cell, atoms2.cell)
    assert np.allclose(atoms1.get_masses(), atoms2.get_masses())
    assert np.allclose(atoms1.get_momenta(), atoms2.get_momenta())

    assert np.allclose(atoms1.get_masses(), atoms2.get_masses())
    assert np.allclose(atoms1.get_initial_charges(), atoms2.get_initial_charges())
    assert np.allclose(atoms1.get_initial_magnetic_moments(), atoms2.get_initial_magnetic_moments())


@pytest.mark.parametrize("atoms", make_atoms_list())
def test_atoms_model(atoms: ase.Atoms, session):
    model = AtomsModel.from_atoms(atoms)
    assert "AtomsModel" in repr(model)
    session.add(model)
    session.commit()

    id = model.id
    assert isinstance(id, int)

    loaded = session.query(AtomsModel).filter_by(id=id).scalar()
    assert isinstance(loaded, AtomsModel)
    loaded_atoms = loaded.to_atoms()
    cmp_atoms(atoms, loaded_atoms)


def test_multiple_atoms(session):
    atoms_list = make_atoms_list()
    for atoms in atoms_list:
        model = AtomsModel.from_atoms(atoms)
        session.add(model)
    session.commit()

    loaded_all = session.query(AtomsModel).all()
    assert len(loaded_all) == len(atoms_list)
    for ii, loaded in enumerate(loaded_all):
        cmp_atoms(loaded.to_atoms(), atoms_list[ii])


def test_atoms_with_calc(session):
    atoms_list = make_atoms_list()

    for expect_id, atoms in enumerate(atoms_list, start=1):
        energy = random.uniform(-10, 10)
        forces = np.random.uniform(-3, 3, size=(len(atoms), 3))
        stress = np.random.uniform(-11, 11, size=(6,))
        magmoms = np.random.uniform(0, 5, size=len(atoms))
        calc = SinglePointCalculator(
            atoms, energy=energy, forces=forces, stress=stress, magmoms=magmoms
        )
        atoms.calc = calc

        session.add(AtomsModel.from_atoms(atoms))
        session.commit()

        del calc
        del atoms

        loaded = session.query(AtomsModel).filter_by(id=expect_id).scalar()
        loaded_atoms = loaded.to_atoms()

        assert pytest.approx(loaded_atoms.get_potential_energy()) == energy
        assert np.allclose(loaded_atoms.get_forces(), forces)
        assert loaded.calculation.fmax == pytest.approx(np.linalg.norm(forces, axis=1).max())
    all_loads = session.query(AtomsModel).all()
    assert len(all_loads) == len(atoms_list)
    all_calcs = session.query(Calculation).all()
    assert len(all_calcs) == len(atoms_list)


def test_query_atoms(session):
    atoms_list = make_atoms_list()
    for atoms in atoms_list:
        session.add(AtomsModel.from_atoms(atoms))
    session.commit()

    found = (
        session.query(AtomsModel)
        .join(Element)
        .where((Element.symbol == "H") & (Element.count == 2))
        .all()
    )
    assert len(found) == 1
    cmp_atoms(atoms_list[0], found[0].to_atoms())

    found = session.query(AtomsModel).join(Element).where(Element.count > 2).all()
    assert len(found) == len(atoms_list) - 1


def test_encode_decode_pbc():
    seen = set()
    for pbc in itertools.product([True, False], [True, False], [True, False]):
        pbc_np = np.array(pbc, dtype=bool)
        val = _encode_pbc(pbc_np)
        assert val not in seen
        seen.add(val)
        assert (pbc_np == _decode_pbc(val)).all()
    assert len(seen) == 2**3


def test_doc_example(session):
    atoms = ase.Atoms("H2O", positions=[[0, 0, 0], [0, 0, 1], [1, 0, 0]])
    atoms_model = AtomsModel.from_atoms(atoms)
    session.add(atoms_model)
    session.commit()

    atoms = atoms_model.to_atoms()
    atoms += ase.Atom("O", position=[1.2, 0, 0])
    atoms_model.set_atoms(atoms)
    session.add(atoms_model)
    session.commit()

    loaded_models = session.query(AtomsModel).all()
    assert len(loaded_models) == 1
    loaded = loaded_models[0]
    elem_counts = loaded.get_element_counts()
    assert len(elem_counts) == 2
    assert elem_counts["H"] == 2
    assert elem_counts["O"] == 2


@pytest.mark.parametrize("atoms", make_atoms_list())
def test_atoms_model_arrays(atoms: ase.Atoms, session):
    atoms_orig = atoms.copy()
    for _ in range(3):
        atoms = atoms_orig.copy()
        numbers = np.random.randint(1, 90, len(atoms))
        atoms.numbers = numbers

        tags = np.random.randint(0, 11, size=len(atoms))
        atoms.set_tags(tags)

        charges = np.random.uniform(-2, 2, size=len(atoms))
        atoms.set_initial_charges(charges)

        ini_magmoms = np.random.uniform(-3, 3, size=len(atoms))
        atoms.set_initial_magnetic_moments(ini_magmoms)

        masses = np.random.uniform(1, 20, size=len(atoms))
        atoms.set_masses(masses)

        momenta = np.random.uniform(-10, 10, size=(len(atoms), 3))
        atoms.set_momenta(momenta)

        model = AtomsModel.from_atoms(atoms)
        session.add(model)
        session.commit()

        id = model.id
        assert isinstance(id, int)
        del model

        loaded_model = session.query(AtomsModel).filter_by(id=id).scalar()
        assert isinstance(loaded_model, AtomsModel)
        loaded = loaded_model.to_atoms()

        assert np.allclose(loaded.get_tags(), tags)
        assert np.allclose(loaded.get_initial_charges(), charges)
        assert np.allclose(loaded.get_initial_magnetic_moments(), ini_magmoms)
        assert np.allclose(loaded.get_momenta(), momenta)
        assert np.allclose(loaded.get_masses(), masses)
        assert np.equal(loaded.numbers, numbers).all()

        cmp_atoms(atoms, loaded)



from __future__ import annotations

import io
import json
from typing import Any, TypeVar

import numpy as np
import sqlalchemy as sa
from ase.cell import Cell
from sqlalchemy.orm import Mapped, declarative_base, mapped_column

from asedb.time_utils import get_posix_timestamp

T_Arr = TypeVar("T_Arr", bound="NamedArray")

ALLOW_PICKLE = False


Base = declarative_base(metadata=sa.MetaData(schema="asedb"))


class ArrayType(sa.types.TypeDecorator):
    """Custom type for saving/loading NumPy arrays."""

    impl = sa.LargeBinary

    def process_bind_param(self, value, dialect):
        """Convert the array going into the database."""
        if value is not None:
            if not isinstance(value, np.ndarray):
                raise TypeError(f"value must be numpy.ndarray, got {value!r}")
            return _serialize_array(value)
        return value

    def process_result_value(self, value, dialect):
        """Convert the result coming from the database."""
        if value is not None:
            value = _deserialize_array(value)
        return value

    def compare_values(self, x: Any, y: Any) -> bool:
        if x is None and y is None:
            return True
        if x is None or y is None:
            # Only 1 value is None
            return False
        if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
            if x.shape != y.shape:
                return False
            if x.dtype != y.dtype:
                return False
            return np.equal(x, y)
        return super().compare_values(x, y)
        # raise TypeError(f"x and y must be None or NumPy arrays, got {x!r} and {y!r}")


class NamedArray(Base):
    """Base table object for storing a NumPy array along with a name and some metadata."""

    __abstract__ = True

    id: Mapped[int] = mapped_column(
        sa.BigInteger().with_variant(sa.Integer, "sqlite"),
        primary_key=True,
    )

    name: Mapped[str] = mapped_column(nullable=False, index=True)

    array_meta_json: Mapped[str] = mapped_column(nullable=True)
    array_obj: Mapped[np.ndarray] = mapped_column(ArrayType, nullable=True)
    last_update_time: Mapped[float] = mapped_column(default=get_posix_timestamp, nullable=False)

    def get_array(self) -> np.ndarray:
        """Access the NumPy array object from the model."""
        return self.array_obj

    @classmethod
    def from_np_array(cls: type[T_Arr], name: str, array: np.ndarray) -> T_Arr:
        """Construct an instance of the NamedArray model from a NumPy array."""
        instance = cls(name=name)
        instance.set_array(array)
        return instance

    def set_array(self, array: np.ndarray | Cell) -> None:
        """Update the blob representing a NumPy array. The array
        will be serialized to a binary blob using the NumPy save function.

        Note: Arbitrary Python objects are not allowed, as Pickle serialization is
        disabled by default for security purposes. Change the asedb.abstract.ALLOW_PICKLE
        variable to True to allow pickle serialization.
        """
        if isinstance(array, Cell):
            # Adapt the ASE Cell object
            array = array.array
        if not isinstance(array, np.ndarray):
            raise TypeError(f"Expected a NumPy array, got {array!r}")
        self.array_meta_json = json.dumps(_get_array_metadata(array))
        self.array_obj = array
        self.last_update_time = get_posix_timestamp()

    @property
    def array_meta(self) -> None | dict[str, Any]:
        if self.array_meta_json is None:
            return None
        return json.loads(self.array_meta_json)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        kwargs = {
            "id": self.id,
            "name": self.name,
        }
        if (meta := self.array_meta) is not None:
            kwargs.update(**meta)
        s = ", ".join([f"{k}='{v}'" for k, v in kwargs.items()])
        return f"{cls_name}({s})"


def _serialize_array(array: np.ndarray) -> bytes:
    """Get the array serialized in bytes."""
    memfile = io.BytesIO()
    np.save(memfile, array, allow_pickle=ALLOW_PICKLE)
    return memfile.getvalue()


def _deserialize_array(blob: bytes) -> np.ndarray:
    """Load the NumPy array from the serialized bytes."""
    memfile = io.BytesIO(blob)
    return np.load(memfile, allow_pickle=ALLOW_PICKLE)


def _get_array_metadata(array: np.ndarray) -> dict[str, Any]:
    """Extract some metadata about the array."""
    meta = {
        "size": array.size,
        "shape": array.shape,
        "dtype": array.dtype.str,
    }
    return meta



from __future__ import annotations

import enum
from collections.abc import Iterator


class _PropertyBase(str, enum.Enum):
    @classmethod
    def iter(cls) -> Iterator[str]:
        """An iterator over the string values of the enum.

        Yields:
            Iterator[str]: The string value of each property.
        """
        for entry in cls:
            yield entry.value

    def __str__(self) -> str:
        """Ensure calling str(...) on the enum returns the string value."""
        return self.value


class ValueProperties(_PropertyBase):
    ENERGY = "energy"
    FREE_ENERGY = "free_energy"
    MAGMOM = "magmom"


class ArrayProperties(_PropertyBase):
    FORCES = "forces"
    STRESS = "stress"
    STRESSES = "stresses"
    CHARGES = "charges"
    MAGMOMS = "magmoms"



from __future__ import annotations

import ase
import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column, relationship

from asedb.abstract import Base
from asedb.atoms_model import CASCADE_DELETE_ALL, AtomsModel

atoms_trajectory_mapping = sa.Table(
    "atoms_trajectory_mapping",
    Base.metadata,
    sa.Column("atoms_id", sa.Integer, sa.ForeignKey("atoms.id")),
    sa.Column("trajectory_id", sa.Integer, sa.ForeignKey("trajectory.id")),
    sa.UniqueConstraint("trajectory_id", "atoms_id"),
)


class Trajectory(Base):
    __tablename__ = "trajectory"
    id: Mapped[int] = mapped_column(primary_key=True)

    project: Mapped[str] = mapped_column(nullable=True, index=True)
    atoms_list: Mapped[list[AtomsModel]] = relationship(
        secondary=atoms_trajectory_mapping,
        cascade=CASCADE_DELETE_ALL,
        single_parent=True,  # 1 atoms object can only belong to 1 trajectory. TODO: Is this what we want?
    )

    def add_atoms(self, atoms: ase.Atoms) -> AtomsModel:
        """Add an Atoms object to the Trajectory.
        Returns the newly created AtomsModel instance, which belongs to the trajectory.
        """
        model = AtomsModel.from_atoms(atoms)
        self.atoms_list.append(model)
        return model

    def to_atoms_list(self) -> list[ase.Atoms]:
        return [model.to_atoms() for model in self.atoms_list]



from __future__ import annotations

import numpy as np
import pytest
from asedb.abstract import _deserialize_array, _serialize_array


@pytest.mark.parametrize(
    "arr",
    [
        np.array([1, 2, 3], dtype=int),
        np.array(["Hello", "World"]),
        np.arange(12).reshape(6, 2).astype(np.float32),
        np.array(["a", "B", "c", "æ", "Åøæå"], dtype=str),
        np.array([1.23, 4.321, 3.1415], dtype=np.float32),
        np.array([1.23, 4.32100001, 3.14159], dtype=np.float64),
        np.array([1.23, 2.0 + 4.12j, 5.11j], dtype=np.complex_),
    ],
)
def test_serialize_array(arr: np.ndarray):
    serialized = _serialize_array(arr)
    assert isinstance(serialized, bytes)

    loaded = _deserialize_array(serialized)
    assert loaded.shape == arr.shape
    assert loaded.dtype == arr.dtype
    assert np.equal(loaded, arr).all()


@pytest.mark.parametrize(
    "shape",
    [
        (10,),
        (15, 15),
        (1, 19),
        (9, 4, 3, 1, 8, 11),
    ],
)
def test_random_arrays(shape: tuple[int, ...]):
    arr = np.random.uniform(-10, 10, size=shape)
    serialized = _serialize_array(arr)
    loaded = _deserialize_array(serialized)

    assert loaded.shape == arr.shape
    assert loaded.dtype == arr.dtype
    assert np.equal(loaded, arr).all()


def test_no_arbitrary_objects():
    arr = np.array([object()])
    assert arr.dtype == np.dtype("object")
    with pytest.raises(ValueError):
        _serialize_array(arr)
